@@ -24,8 +24,6 @@
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
-#include "config.hpp"
-
 #include <fcntl.h>
 #include <iomanip>
 #include <random>
@@ -36,16 +34,26 @@
 #include <boost/program_options.hpp>
 #include <boost/algorithm/string.hpp>
 
+#include <zeep/http/daemon.hpp>
+#include <zeep/http/server.hpp>
+#include <zeep/http/html-controller.hpp>
+#include <zeep/http/rest-controller.hpp>
+#include <zeep/crypto.hpp>
+
 #include "cif++/Secondary.hpp"
 // #include "cif++/Statistics.hpp"
 #include "cif++/CifUtils.hpp"
 #include "cif++/Cif++.hpp"
 #include "cif++/Structure.hpp"
+#include "cif++/Compound.hpp"
 
 #include <zeep/json/element.hpp>
 
-#if USE_RSRC
-#include "mrsrc.hpp"
+#ifdef _MSC_VER
+//MSVC stdlib.h definitions
+#define STDIN_FILENO 0
+#define STDOUT_FILENO 1
+#define STDERR_FILENO 2
 #endif
 
 namespace po = boost::program_options;
@@ -243,12 +251,15 @@ void CompressSimpleArraySelector(OBitStream& inBits, const std::vector<uint32_t>
 				case 4:
 					fits = fits and bn[3] <= w;
 					waste += w - bn[3];
+					[[fallthrough]];
 				case 3:
 					fits = fits and bn[2] <= w;
 					waste += w - bn[2];
+					[[fallthrough]];
 				case 2:
 					fits = fits and bn[1] <= w;
 					waste += w - bn[1];
+					[[fallthrough]];
 				case 1:
 					fits = fits and bn[0] <= w;
 					waste += w - bn[0];
@@ -796,28 +807,17 @@ void DataTable::load(const char* name, std::vector<Data>& table, float& mean, fl
 {
 	using namespace std::literals;
 
-	const float* fv = nullptr;
+	auto rfd = cif::loadResource(name);
 
-#if USE_RSRC
-	mrsrc::rsrc rd(name);
-
-	if (not rd)
+	if (not rfd)
 		throw std::runtime_error("Missing resource "s + name);
 
-	fv = reinterpret_cast<const float*>(rd.data());
-#else
-	std::ifstream rdf(fs::path(DATADIR) / name, std::ios::binary);
-	if (not rdf.is_open())
-		throw std::runtime_error("Missing data file "s + name);
-	
-	rdf.seekg(0, rdf.end);
-	auto size = rdf.tellg();
-	rdf.seekg(0, rdf.beg);
+	rfd->seekg(0, rfd->end);
+	auto size = rfd->tellg();
+	rfd->seekg(0, rfd->beg);
 
-	fv = new float[size / sizeof(float) + 1];
-	rdf.read(reinterpret_cast<char*>(const_cast<float*>(fv)), size);
-	rdf.close();
-#endif
+	const float* fv = new float[size / sizeof(float) + 1];
+	rfd->read(reinterpret_cast<char*>(const_cast<float*>(fv)), size);
 
 	mean = fv[0];
 	sd = fv[1];
@@ -1009,10 +1009,227 @@ json calculateZScores(const Structure& structure)
 	};
 }
 
+// --------------------------------------------------------------------
+#if WEBSERVICE
+
+class tortoize_html_controller : public zeep::http::html_controller
+{
+  public:
+	tortoize_html_controller()
+		: zeep::http::html_controller("tortoize")
+	{
+		mount("{css,scripts,fonts,images,favicon}/", &tortoize_html_controller::handle_file);
+		mount("{favicon.ico,browserconfig.xml,manifest.json}", &tortoize_html_controller::handle_file);
+		mount("", &tortoize_html_controller::index);
+	}
+
+	void index(const zeep::http::request& request, const zeep::http::scope& scope, zeep::http::reply& reply)
+	{
+		get_template_processor().create_reply_from_template("index", scope, reply);
+	}
+};
+
+// --------------------------------------------------------------------
+
+class tortoize_rest_controller : public zeep::http::rest_controller
+{
+  public:
+	tortoize_rest_controller()
+		: zeep::http::rest_controller("")
+		, m_tempdir(fs::temp_directory_path() / "tortoize-ws")
+	{
+		fs::create_directories(m_tempdir);
+
+		map_post_request("tortoize", &tortoize_rest_controller::calculate, "data", "dict");
+	}
+
+	json calculate(const std::string& file, const std::string& dict)
+	{
+		// First store dictionary, just in case
+
+		fs::path dictFile;
+
+		if (not dict.empty())\
+		{
+			dictFile = m_tempdir / ("dict-" + std::to_string(m_next_dict_nr++));
+			std::ofstream tmpFile(dictFile);
+			tmpFile << dict;
+
+			mmcif::CompoundFactory::instance().pushDictionary(dictFile);
+		}
+
+		try
+		{
+			// --------------------------------------------------------------------
+			
+			json data{
+				{ "software",
+					{
+						{ "name", "tortoize" },
+						{ "version", VERSION_STRING },
+						{ "reference", "Sobolev et al. A Global Ramachandran Score Identifies Protein Structures with Unlikely Stereochemistry, Structure (2020)" },
+						{ "reference-doi", "https://doi.org/10.1016/j.str.2020.08.005" }
+					}
+				}
+			};
+
+			// --------------------------------------------------------------------
+
+			mmcif::File f(file.data(), file.length());
+
+			std::set<uint32_t> models;
+			for (auto r: f.data()["atom_site"])
+			{
+				if (not r["pdbx_PDB_model_num"].empty())
+					models.insert(r["pdbx_PDB_model_num"].as<uint32_t>());
+			}
+
+			if (models.empty())
+				models.insert(0);
+
+			for (auto model: models)
+			{
+				Structure structure(f, model);
+				data["model"][std::to_string(model)] = calculateZScores(structure);
+			}
+
+			if (not dictFile.empty())
+			{
+				mmcif::CompoundFactory::instance().popDictionary();
+				fs::remove(dictFile);
+			}
+
+			return data;
+		}
+		catch (...)
+		{
+			std::error_code ec;
+
+			if (not dictFile.empty())
+			{
+				mmcif::CompoundFactory::instance().popDictionary();
+				fs::remove(dictFile);
+			}
+
+			throw;
+		}
+	}
+
+	fs::path m_tempdir;
+	size_t m_next_dict_nr = 1;
+};
+
+int start_server(int argc, char* argv[])
+{
+	using namespace std::literals;
+	namespace zh = zeep::http;
+
+	mmcif::CompoundFactory::init(true);
+
+	int result = 0;
+
+	po::options_description visible_options(PACKAGE_NAME " [options] input [output]");
+	visible_options.add_options()
+		("log",		po::value<std::string>(),	"Write log to this file")
+		
+		("help,h",								"Display help message")
+		("version",								"Print version")
+
+		("address",	po::value<std::string>()->default_value("0.0.0.0"),		"External address")
+		("port",	po::value<uint16_t>()->default_value(10350),			"Port to listen to")
+		("user,u",	po::value<std::string>()->default_value("www-data"),	"User to run the daemon")
+
+		("no-daemon,F",							"Do not fork into background" )
+
+		("verbose,v",							"verbose output")
+		;
+	
+	po::options_description hidden_options("hidden options");
+	hidden_options.add_options()
+		("command", po::value<std::string>(),	"The command to execute")
+		("debug,d",	po::value<int>(),			"Debug level (for even more verbose output)")
+		;
+
+	po::options_description cmdline_options;
+	cmdline_options.add(visible_options).add(hidden_options);
+
+	po::positional_options_description p;
+	p.add("command", 1);
+	
+	po::variables_map vm;
+	po::store(po::command_line_parser(argc, argv).options(cmdline_options).positional(p).run(), vm);
+	
+	po::notify(vm);
+	
+	// --------------------------------------------------------------------
+	
+	std::string secret;
+	if (vm.count("secret"))
+		secret = vm["secret"].as<std::string>();
+	else
+	{
+		secret = zeep::encode_base64(zeep::random_hash());
+		std::cerr << "starting with created secret " << secret << std::endl;
+	}
+
+	std::string user = vm["user"].as<std::string>();
+	std::string address = vm["address"].as<std::string>();
+	uint16_t port = vm["port"].as<uint16_t>();
+
+	zh::daemon server([&]()
+	{
+		auto s = new zeep::http::server();
+
+#if DEBUG
+		s->set_template_processor(new zeep::http::file_based_html_template_processor("docroot"));
+#else
+		s->set_template_processor(new zeep::http::rsrc_based_html_template_processor());
+#endif
+		s->add_controller(new tortoize_rest_controller());
+		s->add_controller(new tortoize_html_controller());
+		return s;
+	}, PACKAGE_NAME );
+
+	std::string command = vm["command"].as<std::string>();
+
+	if (command == "start")
+	{
+		std::cout << "starting server at http://" << address << ':' << port << '/' << std::endl;
+
+		if (vm.count("no-daemon"))
+			result = server.run_foreground(address, port);
+		else
+			result = server.start(address, port, 2, 2, user);
+		// server.start(vm.count("no-daemon"), address, port, 2, user);
+		// // result = daemon::start(vm.count("no-daemon"), port, user);
+	}
+	else if (command == "stop")
+		result = server.stop();
+	else if (command == "status")
+		result = server.status();
+	else if (command == "reload")
+		result = server.reload();
+	else
+	{
+		std::cerr << "Invalid command" << std::endl;
+		result = 1;
+	}
+
+	return result;
+}
+#endif
+
 // --------------------------------------------------------------------
 
 int pr_main(int argc, char* argv[])
 {
+	using namespace std::literals;
+
+#if WEBSERVICE
+	if (argc > 2 and argv[1] == "server"s)
+		return start_server(argc - 1, argv + 1);
+#endif
+
 	po::options_description visible_options(fs::path(argv[0]).filename().string() + " [options] input [output]");
 	visible_options.add_options()
 		("log",		po::value<std::string>(),	"Write log to this file")
